﻿using Kivii.Finances.Entities;
using Kivii.Linq;
using System;
using System.Collections.Generic;
using System.Data;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Kivii.Finances
{
    public static class InvoiceApplyExtension
    {
        /// <summary>
        /// 获取子发票的流水号
        /// </summary>
        /// <param name="invoice"></param>
        /// <param name="conn"></param>
        /// <returns></returns>
        public static string GetSubSerialNumber(this Invoice invoice, IDbConnection conn = null)
        {
            invoice.RootKvid.ThrowIfEmpty("RootKvid不能为空");
            if (conn == null) conn = KiviiContext.GetOpenedDbConnection<Invoice>();
            var rootInvoice = conn.SingleById<Invoice>(invoice.RootKvid);
            rootInvoice.ThrowIfNull("主发票记录不存在");
            var allcount = conn.Count<Invoice>(o => o.RootKvid == rootInvoice.Kvid);
            //allcount不需要加1，因为Root的也统计进去了
            return $"{rootInvoice.SerialNumber}-{allcount}";
        }

        /// <summary>
        /// Payment与Invoice产生关联的方法
        /// </summary>
        /// <param name="invoices"></param>
        /// <param name="payments"></param>
        /// <returns></returns>
        public static List<Invoice> Correlating(this List<Invoice> invoices, List<Payment> payments)
        {
            payments.ThrowIfNullOrEmpty("要关联的Payments不能为空!");
            invoices.ThrowIfNullOrEmpty("要关联的Invoices不能为空!");
            (payments.Exists(o => o.Type != PaymentType.Bank && o.Type != PaymentType.AliPay && o.Type != PaymentType.Cash && o.Type != PaymentType.Pos && o.Type != PaymentType.Split && o.Type != PaymentType.WeChat)).ThrowIfTrue("请选择正确收款进行关联");
            var payedTime = payments.Max(o => o.OperateTime);

            var rtns = new List<Invoice>();

            foreach (var payment in payments)
            {
                //说明已经建立关联关系的payment 跳过
                if (payment.AmountInvoice >= payment.Amount) continue;
                foreach (var invoice in invoices)
                {
                    var amountLeave = payment.Amount - payment.AmountInvoice;//可以开票的金额
                    if (amountLeave <= 0) break;
                    var amountCorrelating = invoice.Amount - invoice.AmountPayment;//要关联的金额
                    if (amountCorrelating <= 0) continue;
                    var amount = amountLeave <= amountCorrelating ? amountLeave : amountCorrelating;

                    invoice.PayedTime = payedTime;
                    invoice.AddOnlyProperties(o => o.PayedTime);
                    invoice.AmountPayment += amount;
                    invoice.AddOnlyProperties(o => o.AmountPayment);

                    #region 建立关联关系
                    var relationInvoice = new Invoice();
                    relationInvoice.PopulateInstance(invoice);
                    relationInvoice.SerialNumber = string.Empty;
                    relationInvoice.Kvid = Guid.NewGuid();
                    relationInvoice.ParentKvid = invoice.Kvid;

                    relationInvoice.BizId = payment.SerialNumber;
                    relationInvoice.BizKvid = payment.RootKvid;//记录根的Kvid
                    relationInvoice.BizType = typeof(Payment).FullName;

                    relationInvoice.Type = "Relation";

                    relationInvoice.PayedTime = payedTime;
                    relationInvoice.Amount = amount;
                    relationInvoice.AmountPayment = amount;
                    //不含税金额
                    relationInvoice.AmountUntaxed = Math.Round(relationInvoice.Amount / (1 + relationInvoice.TaxRate), 2);
                    //税额
                    relationInvoice.AmountTax = relationInvoice.Amount - relationInvoice.AmountUntaxed;
                    relationInvoice.Summary = payment.PayerName;
                    rtns.Add(relationInvoice);
                    #endregion

                    payment.AmountInvoice += amount;
                }
                payment.AddOnlyProperties(o => o.AmountInvoice);
            }

            return rtns;
        }

        /// <summary>
        /// 发票明细拆分
        /// </summary>
        /// <param name="invoiceApplyDetail"></param>
        /// <param name="limit">拆分临界值</param>
        /// <returns></returns>
        internal static List<InvoiceApplyDetail> SplitApplyDetail(this InvoiceApplyDetail invoiceApplyDetail, decimal limit)
        {
            List<InvoiceApplyDetail> applyDetails = new List<InvoiceApplyDetail>();
            #region 金额超过10万需要分开
            if (invoiceApplyDetail.Amount <= limit)
            {
                invoiceApplyDetail.GoodsId = invoiceApplyDetail.GoodsId.IsNullOrEmpty() ? "组1" : invoiceApplyDetail.GoodsId;
                //不含税金额
                invoiceApplyDetail.AmountUntaxed = Math.Round(invoiceApplyDetail.Amount / (1 + invoiceApplyDetail.TaxRate), 2);
                //税额
                invoiceApplyDetail.AmountTax = invoiceApplyDetail.Amount - invoiceApplyDetail.AmountUntaxed;
                //不含税单价
                invoiceApplyDetail.QuantityUnitPriceUntaxed = Math.Round(invoiceApplyDetail.AmountUntaxed / invoiceApplyDetail.Quantity, 4);
                //含税单价
                invoiceApplyDetail.QuantityUnitPrice = Math.Round(invoiceApplyDetail.Amount / invoiceApplyDetail.Quantity, 4);
                applyDetails.Add(invoiceApplyDetail);
            }
            else
            {
                var count = (int)(invoiceApplyDetail.Amount / limit);
                for (var i = 0; i <= count - 1; i++)
                {
                    var detail = new InvoiceApplyDetail();

                    detail.GoodsId = $"发票{i}";
                    detail.GoodsFullName = invoiceApplyDetail.GoodsFullName;
                    detail.GoodsSpecifications = invoiceApplyDetail.GoodsSpecifications;
                    detail.QuantityUnit = invoiceApplyDetail.QuantityUnit;
                    detail.QuantityUnitPriceUntaxed = invoiceApplyDetail.QuantityUnitPriceUntaxed;
                    detail.Quantity = invoiceApplyDetail.Quantity;
                    detail.TaxRate = invoiceApplyDetail.TaxRate;

                    detail.Amount = limit;
                    //不含税金额
                    detail.AmountUntaxed = Math.Round(detail.Amount / (1 + detail.TaxRate), 2);
                    //税额
                    detail.AmountTax = detail.Amount - detail.AmountUntaxed;
                    //不含税单价
                    detail.QuantityUnitPriceUntaxed = Math.Round(detail.AmountUntaxed / detail.Quantity, 4);
                    //含税单价
                    detail.QuantityUnitPrice = Math.Round(detail.Amount / detail.Quantity, 4);

                    applyDetails.Add(detail);
                }
                var leftAmount = invoiceApplyDetail.Amount - limit * count;
                if (leftAmount > 0)
                {
                    var item = new InvoiceApplyDetail();
                    item.GoodsId = $"发票{count}";
                    item.GoodsFullName = invoiceApplyDetail.GoodsFullName;
                    item.GoodsSpecifications = invoiceApplyDetail.GoodsSpecifications;
                    item.QuantityUnit = invoiceApplyDetail.QuantityUnit;
                    item.QuantityUnitPriceUntaxed = invoiceApplyDetail.QuantityUnitPriceUntaxed;
                    item.Quantity = invoiceApplyDetail.Quantity;
                    item.TaxRate = invoiceApplyDetail.TaxRate;

                    item.Amount = invoiceApplyDetail.Amount - limit * count;
                    item.AmountUntaxed = Math.Round(item.Amount / (1 + item.TaxRate), 2);
                    item.AmountTax = item.Amount - item.AmountUntaxed;
                    item.QuantityUnitPriceUntaxed = Math.Round(item.AmountUntaxed / item.Quantity, 4);
                    item.QuantityUnitPrice = Math.Round(item.Amount / item.Quantity, 4);
                    applyDetails.Add(item);
                }
            }
            #endregion
            return applyDetails;
        }
    }
}
